// Copyright Earl Warren <contact@earl-warren.org>
// Copyright Loïc Dachary <loic@dachary.org>
// SPDX-License-Identifier: MIT

package f3

import (
	"encoding/json"
	"slices"
	"strings"

	"code.forgejo.org/f3/gof3/v3/util"

	"github.com/google/go-cmp/cmp"
)

type Comparer any

type Interface interface {
	GetID() string
	GetName() string
	SetID(id string)
	IsNil() bool
	GetReferences() References
	ToReference() *Reference
	GetCmpIgnoreFields() []string
	GetComparers() []Comparer
	Clone() Interface
}

type ReferenceInterface interface {
	Get() string
	Set(reference string)
	GetIDAsString() string
	GetIDAsInt() int64
}

type Reference struct {
	ID string
}

func (r *Reference) Get() string {
	return r.ID
}

func (r *Reference) Set(reference string) {
	r.ID = reference
}

func (r *Reference) GetIDAsInt() int64 {
	return util.ParseInt(r.GetIDAsString())
}

func (r *Reference) GetIDAsString() string {
	s := strings.Split(r.ID, "/")
	return s[len(s)-1]
}

type References []ReferenceInterface

func (r *Reference) UnmarshalJSON(b []byte) error {
	return json.Unmarshal(b, &r.ID)
}

func (r Reference) MarshalJSON() ([]byte, error) {
	return json.Marshal(r.ID)
}

func (r *Reference) GetID() string   { return r.ID }
func (r *Reference) SetID(id string) { r.ID = id }
func (r *Reference) IsNil() bool     { return r.ID == "0" || r.ID == "" }

func NewReferences() References {
	return make([]ReferenceInterface, 0, 1)
}

func NewReference(id string) *Reference {
	r := &Reference{}
	r.SetID(id)
	return r
}

type Common struct {
	Index Reference `json:"index"`
}

func (c *Common) GetID() string                { return c.Index.GetID() }
func (c *Common) GetName() string              { return c.GetID() }
func (c *Common) SetID(id string)              { c.Index.SetID(id) }
func (c *Common) IsNil() bool                  { return c.Index.IsNil() }
func (c *Common) GetReferences() References    { return NewReferences() }
func (c *Common) ToReference() *Reference      { return &c.Index }
func (c *Common) GetCmpIgnoreFields() []string { return []string{".Index"} }
func (c *Common) GetComparers() []Comparer     { return []Comparer{} }

func newCmpCompareOptions(f Interface) cmp.Options {
	options := make(cmp.Options, 0, 5)
	for _, comparer := range f.GetComparers() {
		options = append(options, cmp.Comparer(comparer))
	}
	return options
}

func NewCmpOptions(f Interface) cmp.Options {
	ignore := f.GetCmpIgnoreFields()
	options := cmp.Options{
		cmp.FilterPath(
			func(path cmp.Path) bool {
				return slices.Contains(ignore, path.Last().String())
			},
			cmp.Ignore()),
	}
	return append(options, newCmpCompareOptions(f)...)
}

var Nil = &Common{}

func NewCommon(id string) Common {
	return Common{Index: *NewReference(id)}
}

func (c *Common) Clone() Interface {
	clone := &Common{}
	*clone = *c
	return clone
}
