// Copyright Earl Warren <contact@earl-warren.org>
// Copyright Loïc Dachary <loic@dachary.org>
// SPDX-License-Identifier: MIT

package f3

import (
	"fmt"
	"slices"

	"code.forgejo.org/f3/gof3/v3/tree/generic"
)

type Path interface {
	NodeIDs() []generic.NodeID
	OwnerAndProjectID() (owner, project int64)

	AppendID(id string) Path
	Ignore() Path

	generic.Path
	Root() Path

	Forge() Path
	SetForge() Path

	Assets() Path
	SetAssets() Path

	Comments() Path
	SetComments() Path

	Issues() Path
	SetIssues() Path

	Labels() Path
	SetLabels() Path

	Milestones() Path
	SetMilestones() Path

	Owners() Path
	SetOwners(owners generic.Kind) Path

	Organizations() Path
	SetOrganizations() Path

	Projects() Path
	SetProjects() Path

	PullRequests() Path
	SetPullRequests() Path

	Reactions() Path
	SetReactions() Path

	Releases() Path
	SetReleases() Path

	Repositories() Path
	SetRepositories() Path

	Reviews() Path
	SetReviews() Path

	ReviewComments() Path
	SetReviewComments() Path

	Topics() Path
	SetTopics() Path

	Users() Path
	SetUsers() Path
}

type path struct {
	generic.PathImplementation
}

func (o path) popKind(kind ...generic.Kind) Path {
	firstKind := generic.Kind(o.First().GetID().String())
	if !slices.Contains(kind, firstKind) {
		panic(fmt.Errorf("%s expected one of %s got %s", o, kind, firstKind))
	}
	return ToPath(o.RemoveFirst())
}

func (o path) NodeIDs() []generic.NodeID {
	nodeIDs := make([]generic.NodeID, 0, o.Length())
	collectID := false
	for _, node := range o.Root().All() {
		if collectID {
			nodeIDs = append(nodeIDs, node.GetID())
			collectID = false
			continue
		}
		kind := generic.Kind(node.GetID().String())
		fixedID, ok := containerChildFixedID[kind]
		if !ok {
			panic(fmt.Errorf("%s unexpected kind %s", o.All(), kind))
		}
		if !fixedID {
			collectID = true
		}
	}
	return nodeIDs
}

func (o path) OwnerAndProjectID() (owner, project int64) {
	nodeIDs := o.NodeIDs()
	return nodeIDs[0].Int64(), nodeIDs[1].Int64()
}

func (o path) Root() Path { return o.popKind(generic.Kind("")) }

func (o path) Forge() Path    { return o.popKind(KindForge) }
func (o path) SetForge() Path { return o.appendKind(KindForge) }

func (o path) Assets() Path    { return o.popKind(KindAssets) }
func (o path) SetAssets() Path { return o.appendKind(KindAssets) }

func (o path) Comments() Path    { return o.popKind(KindComments) }
func (o path) SetComments() Path { return o.appendKind(KindComments) }

func (o path) Issues() Path    { return o.popKind(KindIssues) }
func (o path) SetIssues() Path { return o.appendKind(KindIssues) }

func (o path) Labels() Path    { return o.popKind(KindLabels) }
func (o path) SetLabels() Path { return o.appendKind(KindLabels) }

func (o path) Milestones() Path    { return o.popKind(KindMilestones) }
func (o path) SetMilestones() Path { return o.appendKind(KindMilestones) }

func (o path) Organizations() Path    { return o.popKind(KindOrganizations) }
func (o path) SetOrganizations() Path { return o.appendKind(KindOrganizations) }

func (o path) Projects() Path    { return o.popKind(KindProjects) }
func (o path) SetProjects() Path { return o.appendKind(KindProjects) }

func (o path) PullRequests() Path    { return o.popKind(KindPullRequests) }
func (o path) SetPullRequests() Path { return o.appendKind(KindPullRequests) }

func (o path) Reactions() Path    { return o.popKind(KindReactions) }
func (o path) SetReactions() Path { return o.appendKind(KindReactions) }

func (o path) Releases() Path    { return o.popKind(KindReleases) }
func (o path) SetReleases() Path { return o.appendKind(KindReleases) }

func (o path) Repositories() Path    { return o.popKind(KindRepositories) }
func (o path) SetRepositories() Path { return o.appendKind(KindRepositories) }

func (o path) Reviews() Path    { return o.popKind(KindReviews) }
func (o path) SetReviews() Path { return o.appendKind(KindReviews) }

func (o path) ReviewComments() Path    { return o.popKind(KindReviewComments) }
func (o path) SetReviewComments() Path { return o.appendKind(KindReviewComments) }

func (o path) Topics() Path    { return o.popKind(KindTopics) }
func (o path) SetTopics() Path { return o.appendKind(KindTopics) }

func (o path) Users() Path    { return o.popKind(KindUsers) }
func (o path) SetUsers() Path { return o.appendKind(KindUsers) }

func (o path) Owners() Path                       { return o.popKind(KindUsers, KindOrganizations) }
func (o path) SetOwners(owners generic.Kind) Path { return o.appendKind(owners) }

func (o path) AppendID(id string) Path {
	return ToPath(o.Append(generic.NewNodeFromID(id)))
}

func (o path) appendKind(kind generic.Kind) Path {
	return o.AppendID(string(kind))
}

func (o path) Ignore() Path {
	return ToPath(o.RemoveFirst())
}

func ToPath(other generic.Path) Path {
	return path{other.(generic.PathImplementation)}
}

func NewPathFromString(pathString string) Path {
	return ToPath(generic.NewPathFromString(pathString))
}

func NewPath(nodes ...generic.NodeInterface) Path {
	return path{
		PathImplementation: generic.PathImplementation(nodes),
	}
}
